//! Known qualifier types for [gem].
//!
//! [gem]: https://github.com/package-url/purl-spec/blob/master/PURL-TYPES.rst#gem

use super::str_ref_qualifier;

str_ref_qualifier!(Platform, "platform", "platform");
